"""Module to return information and entropy from TFR Power"""

import numpy as np
from scipy import signal
import quantum_inferno.scales_dyadic as scales
from quantum_inferno import utils
import scipy.fft


# TODO: Use 'em or lose 'em
def log2_ceil(x: float, epsilon: float = scales.EPSILON64) -> float:
    """
    Compute ceiling of log2 of a positive input argument.
    Corrects for negative, complex or zero inputs by
    taking the absolute value and adding EPSILON
    :param x: input, converts to positive real
    :param epsilon: override zero, negative, and imaginary values
    :return: ceiling of log2
    """
    real_x = np.abs(x) + epsilon
    return np.ceil(np.log2(real_x))


def log2_round(x: float, epsilon: float = scales.EPSILON64) -> float:
    """
    Compute ceiling of log2 of a positive input argument.
    Corrects for negative, complex or zero inputs by
    taking the absolute value and adding EPSILON
    :param x: input, converts to positive real
    :param epsilon: override zero, negative, and imaginary values
    :return: ceiling of log2
    """
    real_x = np.abs(x) + epsilon
    return np.round(np.log2(real_x))


def log2_floor(x: float, epsilon: float = scales.EPSILON64) -> float:
    """
    Compute ceiling of log2 of a positive input argument.
    Corrects for negative, complex or zero inputs by
    taking the absolute value and adding EPSILON
    :param x: input, converts to positive real
    :param epsilon: override zero, negative, and imaginary values
    :return: ceiling of log2
    """
    real_x = np.abs(x) + epsilon
    return np.floor(np.log2(real_x))


def mat_max_idx(a: np.ndarray):
    """Find the indexes of the max of a matrix"""
    return np.unravel_index(a.argmax(), a.shape)


def mat_min_idx(a: np.ndarray):
    """Find the indexes of the min of a matrix"""
    return np.unravel_index(a.argmin(), a.shape)


def power_dynamics_scaled_bits(tfr_power: np.ndarray) -> [np.ndarray, np.ndarray, np.ndarray]:
    """
    Essential scales for power
    :param tfr_power: power from time-frequency representation
    :return:
    """
    tfr_power_bits = np.log2(tfr_power + scales.EPSILON64)
    tfr_power_bits -= np.max(tfr_power_bits)
    # Dynamic range per time step re max
    tfr_power_per_time = np.sum(tfr_power, axis=0)
    tfr_power_per_time_bits = np.log2(tfr_power_per_time + scales.EPSILON64)
    tfr_power_per_time_bits -= np.max(tfr_power_per_time_bits)
    # Dynamic range per frequency band re max
    tfr_power_per_freq = np.sum(tfr_power, axis=1)
    tfr_power_per_freq_bits = np.log2(tfr_power_per_freq + scales.EPSILON64)
    tfr_power_per_freq_bits -= np.max(tfr_power_per_freq_bits)

    return [tfr_power_bits, tfr_power_per_freq_bits, tfr_power_per_time_bits]


def shannon(tfr_power):
    """
    Shannon information and entropy
    :param tfr_power:
    :return:
    """
    atom_bits = 3
    num_freq = tfr_power.shape[0]
    num_time = tfr_power.shape[1]
    num_dof = num_freq*num_time  # Degrees of freedom
    ref_shannon_bits = np.log2(num_dof)/num_dof
    # ref_shannon_bits = (np.log2(num_freq)+3)/num_dof

    # Normalized power pdf, information, and entropy over whole distribution
    tfr_power_pdf = tfr_power/np.sum(tfr_power)
    # Shannon information; NOT ADDITIVE, but already normalized
    tfr_info = -np.log2(tfr_power_pdf + scales.EPSILON64)
    tfr_isnr = np.log2(num_dof) - tfr_info
    # Shannon entropy per bin
    # This is the expected value of the information
    tfr_shannon_bits = tfr_power_pdf * tfr_info
    tfr_shannon_total_bits = np.sum(tfr_shannon_bits)
    # Relative to ref_bits
    tfr_esnr = tfr_shannon_bits/ref_shannon_bits

    # # tfr_esnr = num_dof*tfr_shannon_bits/tfr_shannon_total_bits
    # tfr_esnr = (tfr_shannon_bits - ref_shannon_bits)*num_dof
    # tfr_esnr = tfr_shannon_bits/ref_shannon_bits/atom_bits
    # tfr_esnr = (np.log2(num_dof) - tfr_shannon_bits)*num_dof
    # # Relative entropy dsnr
    # tfr_dsnr_uni = ref_shannon_bits-tfr_shannon_bits
    # # Total relative entropy
    # tfr_DSNR_total = num_dof*ref_shannon_bits-tfr_shannon_total_bits

    return [tfr_info, tfr_shannon_bits, tfr_isnr, tfr_esnr]


def shannon_esnrt_per_time(tfr_power):
    """
    Normalized power pdf, information, and entropy per time step
    :param tfr_power:
    :return:
    """

    num_freq = tfr_power.shape[0]
    num_dof = num_freq  # Degrees of freedom
    ref_shannon_bits = np.log2(num_dof)/num_dof

    tfr_power_per_time = np.sum(tfr_power, axis=0) + scales.EPSILON64
    tfr_power_per_time_pdf = utils.d1tile_x_d0d1(d1=1/tfr_power_per_time, d0d1=tfr_power)
    tfr_info_per_time = -np.log2(tfr_power_per_time_pdf + scales.EPSILON64)
    tfr_isnr_per_time = np.log2(num_dof) - tfr_info_per_time

    tfr_shannon_per_time_bits = tfr_power_per_time_pdf * tfr_info_per_time
    # tfr_shannon_sum_per_time_bits = np.sum(tfr_shannon_per_time_bits, axis=0)
    # tfr_esnr_per_time = \
    #     utils.d1tile_x_d0d1(d1=tfr_power.shape[0]/tfr_shannon_sum_per_time_bits, d0d1=tfr_shannon_per_time_bits)
    # Relative to ref_bits
    # tfr_esnr_per_time = (tfr_shannon_per_time_bits - ref_shannon_bits)*num_dof
    tfr_esnr_per_time = tfr_shannon_per_time_bits/ref_shannon_bits

    return [tfr_info_per_time, tfr_shannon_per_time_bits, tfr_isnr_per_time, tfr_esnr_per_time]


def shannon_esnrf_per_freq(tfr_power):
    """
    Normalized power pdf, information, and entropy per frequency step
    :param tfr_power:
    :return:
    """

    atom_bits = 3
    num_time = tfr_power.shape[1]
    num_dof = num_time  # Degrees of freedom
    ref_shannon_bits = np.log2(num_dof)/num_dof
    # ref_shannon_bits = 3/num_dof

    tfr_power_per_freq = np.sum(tfr_power, axis=1) + scales.EPSILON64
    tfr_power_per_freq_pdf = utils.d0tile_x_d0d1(d0=1/tfr_power_per_freq, d0d1=tfr_power)
    tfr_info_per_freq = -np.log2(tfr_power_per_freq_pdf + scales.EPSILON64)
    # TODO: taper the np.log2(num_dof), otherwise blows up at edges
    # TODO: Verify, standardize with dsp taper
    tukey_power_envelope = signal.windows.tukey(M=num_time, alpha=0.2, sym=True)**2
    power_taper_tile = utils.just_tile_d1(d1_array1d_in=tukey_power_envelope, d0d1_shape=tfr_power.shape)
    tfr_isnr_per_freq = np.log2(num_dof*power_taper_tile + scales.EPSILON64) - tfr_info_per_freq

    # # Debug
    # print('\nlog2DOF:', np.log2(num_dof))
    # print(num_dof_tapered.shape)
    # print(tfr_power.shape)
    # print('tfr_power_per_freq:', tfr_power_per_freq)
    # print('tfr_isnr_per_freq[:, -100]:', tfr_isnr_per_freq[:, -100])
    # print('tfr_isnr_per_freq[:, -1]:', tfr_isnr_per_freq[:, -1])
    # exit()
    # TODO: Reconcile with dsp taper
    tfr_shannon_per_freq_bits = tfr_power_per_freq_pdf * power_taper_tile * tfr_info_per_freq
    # tfr_shannon_sum_per_freq_bits = np.sum(tfr_shannon_per_freq_bits, axis=1)
    # tfr_esnr_per_freq = \
    #     utils.d0tile_x_d0d1(d0=tfr_power.shape[1]/tfr_shannon_sum_per_freq_bits, d0d1=tfr_shannon_per_freq_bits)
    # Relative to ref_bits
    # tfr_esnr_per_freq = (tfr_shannon_per_freq_bits - ref_shannon_bits)*num_dof
    # tfr_esnr_per_freq = tfr_shannon_per_freq_bits/ref_shannon_bits
    tfr_esnr_per_freq = tfr_shannon_per_freq_bits/ref_shannon_bits  # /atom_bits

    return [tfr_info_per_freq, tfr_shannon_per_freq_bits, tfr_isnr_per_freq, tfr_esnr_per_freq]


def shannon_tdr_fft(sig_in_real, verbose: bool = True):
    """
    Shannon information and entropy
    :param sig_in_real:
    :param verbose: print to screen
    :return:
    """

    atom_bits = 3
    num_dof = len(sig_in_real)
    ref_shannon_bits = np.log2(num_dof)/num_dof

    # Time-domain representation
    sig_sq_total = np.sum(sig_in_real**2)
    tdr_sig = sig_in_real/np.sqrt(sig_sq_total)
    # Time marginals
    tdr_marginal = tdr_sig**2
    tdr_info = -np.log2(tdr_marginal + scales.EPSILON32)
    tdr_entropy = tdr_marginal*tdr_info
    # tdr_marginal_total = np.sum(tdr_marginal)
    # tdr_entropy_total = np.sum(tdr_entropy)

    # FFT information and entropy of real input signal
    fft_sig = scipy.fft.rfft(x=sig_in_real)
    fft_angle_rads = np.unwrap(np.angle(fft_sig))
    # RFFT only goes up to Nyquist TODO: use rfftfreq, presently assuming sample rate of unity
    fft_frequency = np.arange(len(fft_angle_rads))/len(fft_angle_rads)/2.
    # Frequency marginals
    fft_sq = np.abs(fft_sig)**2
    fft_sq_total = np.sum(fft_sq)
    fft_marginal = fft_sq/fft_sq_total
    fft_info = -np.log2(fft_marginal + scales.EPSILON32)
    fft_entropy = fft_marginal*fft_info
    # fft_marginal_total = np.sum(fft_marginal)
    # fft_entropy_total = np.sum(fft_entropy)

    # Average entropy for P ~ 1/NFFT
    ref_tdr_entropy = np.log2(len(tdr_marginal))/len(tdr_marginal)
    ref_fft_entropy = np.log2(len(fft_marginal))/len(fft_marginal)

    # THE CRUX
    tdr_isnr = np.log2(len(tdr_info)) - tdr_info
    tdr_esnr = tdr_entropy/ref_tdr_entropy
    fft_isnr = np.log2(len(fft_info)) - fft_info
    fft_esnr = fft_entropy/ref_fft_entropy

    if verbose:
        print('Ref entropy, time:', ref_tdr_entropy)
        print('Ref entropy, frequency:', ref_fft_entropy)
        print('Total Entropy, time:', np.sum(tdr_entropy))
        print('Total Entropy, frequency:', 2*np.sum(fft_entropy))
        print('Sum of time marginal:', np.sum(tdr_marginal))
        print('Sum of frequency marginal:', np.sum(fft_marginal))

    return [tdr_sig, tdr_marginal, tdr_info, tdr_entropy, tdr_isnr, tdr_esnr,
            fft_sig, fft_marginal, fft_frequency, fft_angle_rads, fft_info, fft_entropy, fft_isnr, fft_esnr]


def shannon_tdr(sig_in_real, verbose: bool = True):
    """
    Shannon information and entropy
    :param sig_in_real:
    :param verbose: print to screen
    :return:
    """

    # Time-domain representation
    sig_sq_total = np.sum(sig_in_real**2)
    tdr_sig = sig_in_real/np.sqrt(sig_sq_total)
    # Time marginals
    # TODO: rel relative to q=1/M; log2(p/q), after Rosso et al. 2001
    tdr_marginal = tdr_sig**2
    tdr_info = -np.log2(tdr_marginal + scales.EPSILON32)
    tdr_info_rel = np.log2(tdr_marginal*len(tdr_marginal) + scales.EPSILON32)
    tdr_entropy = tdr_marginal*tdr_info
    tdr_entropy_total = np.sum(tdr_entropy)
    # Average entropy for P ~ 1/NFFT
    ref_tdr_entropy = np.log2(len(tdr_marginal))/len(tdr_marginal)
    ref_tdr_rel_entropy = 1/(2*np.log(2))*(np.max(tdr_marginal)-1/len(tdr_marginal))**2

    # THE CRUX
    tdr_isnr = np.log2(len(tdr_info)) - tdr_info
    # Stable
    tdr_esnr = tdr_entropy/ref_tdr_entropy
    # TODO: Experiment
    # tdr_rel_info = np.log2(tdr_marginal*len(tdr_marginal) + scales.EPSILON32)
    # May be promising, see Lin (1999)
    # tdr_rel_info = np.log2(tdr_marginal/(0.5*tdr_marginal + 0.5/len(tdr_marginal)) + scales.EPSILON32)
    # tdr_rel_entropy = tdr_marginal*tdr_rel_info
    # Doesn't go to zero
    # tdr_rel_entropy = (tdr_marginal-1/len(tdr_marginal))*tdr_rel_info
    # # Formal rel entropy
    # tdr_esnr = tdr_rel_entropy
    # Promising
    # tdr_esnr = tdr_rel_entropy/ref_tdr_rel_entropy

    if verbose:
        print('Ref entropy, time:', ref_tdr_entropy)
        print('Total Entropy, time:', np.sum(tdr_entropy))
        print('Sum of time marginal:', np.sum(tdr_marginal))

    return [tdr_sig, tdr_marginal, tdr_info, tdr_entropy, tdr_isnr, tdr_esnr]


def shannon_fft(fft_sig, verbose: bool = True):
    """
    Shannon information and entropy
    :param fft_sig:
    :param verbose: print to screen
    :return:
    """

    # FFT information and entropy of real input signal
    # fft_sig = scipy.fft.rfft(x=sig_in_real)
    fft_angle_rads = np.unwrap(np.angle(fft_sig))
    # # RFFT only goes up to Nyquist
    # fft_frequency = np.arange(len(fft_angle_rads))/len(fft_angle_rads)/2.
    # Frequency marginals
    fft_sq = np.abs(fft_sig)**2
    fft_sq_total = np.sum(fft_sq)
    fft_marginal = fft_sq/fft_sq_total
    fft_info = -np.log2(fft_marginal + scales.EPSILON32)
    fft_info_rel = np.log2(fft_marginal*len(fft_marginal) + scales.EPSILON32)
    fft_entropy = fft_marginal*fft_info
    # fft_marginal_total = np.sum(fft_marginal)
    fft_entropy_total = np.sum(fft_entropy)

    # Average entropy for P ~ 1/NFFT
    ref_fft_entropy = np.log2(len(fft_marginal)) / len(fft_marginal)
    # ref_fft_rel_entropy = 1/(2*np.log(2))*(np.max(fft_marginal)-1/len(fft_marginal))**2

    # THE CRUX
    fft_isnr = np.log2(len(fft_info)) - fft_info
    # Stable
    fft_esnr = fft_entropy/ref_fft_entropy
    # TODO: Experiments
    # fft_esnr = fft_marginal*fft_info_rel/ref_fft_entropy
    # fft_esnr = (fft_entropy - np.log2(len(fft_marginal)))
    # fft_rel_info = np.log2(fft_marginal*len(fft_marginal) + scales.EPSILON32)
    # fft_rel_entropy = fft_marginal*fft_rel_info
    # Formal def
    # fft_esnr = fft_rel_entropy
    # # Not bad
    # fft_esnr = np.log2(len(fft_marginal))*fft_marginal*fft_rel_info
    # Promising
    # fft_esnr = fft_entropy_total * fft_rel_entropy
    # Large
    # fft_esnr = fft_rel_entropy/ref_fft_rel_entropy
    # This should work
    # fft_esnr = np.log2(len(fft_marginal))/len(fft_marginal) - fft_entropy

    if verbose:
        print('Ref entropy, frequency:', ref_fft_entropy)
        print('Total Entropy, frequency:', 2*np.sum(fft_entropy))
        print('Sum of frequency marginal:', np.sum(fft_marginal))

    return [fft_marginal, fft_angle_rads, fft_info, fft_entropy, fft_isnr, fft_esnr]


def shannon_stft(tfr_power):
    """
    Shannon information and entropy
    :param tfr_power:
    :return:
    """

    atom_bits = 3
    num_freq = tfr_power.shape[0]
    num_time = tfr_power.shape[1]
    # Previous
    num_dof = num_freq*num_time    # Degrees of freedom
    ref_shannon_bits = np.log2(num_dof)/num_dof
    # TODO: Works OK
    # num_dof = num_time*(1-stft_overlap_fractional)    # Degrees of freedom
    # num_dof = num_time    # Degrees of freedom
    # ref_shannon_bits = np.log2(num_dof)/num_dof

    # Normalized power pdf, information, and entropy over whole distribution
    tfr_power_pdf = tfr_power/np.sum(tfr_power)
    # Shannon information; NOT ADDITIVE, but already normalized
    tfr_info = -np.log2(tfr_power_pdf + scales.EPSILON64)
    tfr_isnr = np.log2(num_dof) - tfr_info
    # Shannon entropy per bin
    # This is the expected value of the information
    tfr_shannon_bits = tfr_power_pdf * tfr_info
    # Relative to ref_bits
    tfr_esnr = tfr_shannon_bits/ref_shannon_bits

    # print('NFFT:', nfft)
    # print('Ref shannon bits')
    # print(ref_shannon_bits)
    # print('min max tfr_shannon_bits')
    # print(np.min(tfr_shannon_bits), np.max(tfr_shannon_bits))

    return [tfr_info, tfr_shannon_bits, tfr_isnr, tfr_esnr]


def shannon_stft_esnrt_per_time(tfr_power):
    """
    Normalized power pdf, information, and entropy per time step
    :param tfr_power:
    :return:
    """

    num_time = tfr_power.shape[1]
    num_freq = tfr_power.shape[0]

    # # TODO: Clean up!
    # num_dof = num_time    # Degrees of freedom
    num_dof = num_freq
    ref_shannon_bits = np.log2(num_dof)/num_dof

    tfr_power_per_time = np.sum(tfr_power, axis=0) + scales.EPSILON64
    tfr_power_per_time_pdf = utils.d1tile_x_d0d1(d1=1/tfr_power_per_time, d0d1=tfr_power)
    tfr_info_per_time = -np.log2(tfr_power_per_time_pdf + scales.EPSILON64)
    tfr_isnr_per_time = np.log2(num_freq) - tfr_info_per_time

    tfr_shannon_per_time_bits = tfr_power_per_time_pdf * tfr_info_per_time
    # Relative to ref_bits
    tfr_esnr_per_time = tfr_shannon_per_time_bits/ref_shannon_bits

    print('Ref tfr_shannon_per_time_bits')
    print(ref_shannon_bits)
    print('min max tfr_shannon_per_time_bits')
    print(np.min(tfr_shannon_per_time_bits), np.max(tfr_shannon_per_time_bits))

    return [tfr_info_per_time, tfr_shannon_per_time_bits, tfr_isnr_per_time, tfr_esnr_per_time]


def shannon_stft_esnrf_per_freq(tfr_power):
    """
    Normalized power pdf, information, and entropy per frequency step
    :param tfr_power:
    :return:
    """

    atom_bits = 3
    num_time = tfr_power.shape[1]
    num_freq = tfr_power.shape[0]
    num_dof = num_time    # Degrees of freedom
    ref_shannon_bits = np.log2(num_dof)/num_dof

    tfr_power_per_freq = np.sum(tfr_power, axis=1) + scales.EPSILON64
    # Threshold the power to prevent it from blowing up near Nyquist
    # min_inv = np.min(inv_tfr_power_per_freq)
    # idx = np.argwhere(inv_tfr_power_per_freq > min_inv*2**32)
    # inv_tfr_power_per_freq[idx] = np.zeros(idx.shape)

    tfr_power_per_freq_pdf = utils.d0tile_x_d0d1(d0=1/tfr_power_per_freq, d0d1=tfr_power)
    tfr_info_per_freq = -np.log2(tfr_power_per_freq_pdf + scales.EPSILON64)
    tfr_isnr_per_freq = np.log2(num_dof) - tfr_info_per_freq

    tfr_shannon_per_freq_bits = tfr_power_per_freq_pdf * tfr_info_per_freq

    # Relative to ref_bits
    tfr_esnr_per_freq = tfr_shannon_per_freq_bits/ref_shannon_bits

    print('Ref tfr_esnr_per_freq bits')
    print(ref_shannon_bits)
    print('min max tfr_esnr_per_freq')
    print(np.min(tfr_esnr_per_freq), np.max(tfr_esnr_per_freq))

    return [tfr_info_per_freq, tfr_shannon_per_freq_bits, tfr_isnr_per_freq, tfr_esnr_per_freq]

# def shannon_per_freq_order(tfr_power, order: float = 3):
#     """
#     Normalized power pdf, information, and entropy per frequency step
#     :param order:
#     :param tfr_power:
#     :return:
#     """
#
#     # TODO: Pass the number of wavelets per band
#     num_freq = tfr_power.shape[0]
#     scaling = 2**(np.arange(num_freq)/order)/2
#
#     num_time = tfr_power.shape[1]
#     num_dof = num_time  # Degrees of freedom
#     ref_shannon_bits = np.log2(num_dof)/num_dof
#
#     tfr_power_per_freq = np.sum(tfr_power, axis=1)
#     tfr_power_per_freq_pdf = utils.d0tile_x_d0d1(d0=1/tfr_power_per_freq, d0d1=tfr_power)
#     tfr_info_per_freq = -np.log2(tfr_power_per_freq_pdf + scales.EPSILON64)
#     tfr_isnr_per_freq = np.log2(tfr_power.shape[1]) - tfr_info_per_freq
#
#     tfr_shannon_per_freq_bits = tfr_power_per_freq_pdf * tfr_info_per_freq
#     # tfr_shannon_sum_per_freq_bits = np.sum(tfr_shannon_per_freq_bits, axis=1)
#     # tfr_esnr_per_freq = \
#     #     utils.d0tile_x_d0d1(d0=tfr_power.shape[1]/tfr_shannon_sum_per_freq_bits, d0d1=tfr_shannon_per_freq_bits)
#     # tfr_esnr_per_freq = tfr_shannon_per_freq_bits/ref_shannon_bits
#     tfr_esnr_per_freq = \
#         utils.d0tile_x_d0d1(d0=1/ref_shannon_bits/scaling, d0d1=tfr_shannon_per_freq_bits)
#
#     return [tfr_info_per_freq, tfr_shannon_per_freq_bits, tfr_isnr_per_freq, tfr_esnr_per_freq]


""""
Standardized Stockwell transform (stx) with optimization parameters
After Moukadem et al., 2022, A new optimized Stockwell transform applied on synthetic and real non-stationary signals
Rederivation in preparation, Garces et al. 2022; last updated 9 May 2022

"""
import numpy as np
from scipy.fft import fft, ifft, fftfreq
from quantum_inferno import scales_dyadic as scales
from typing import Tuple


def sig_pad_up_to_pow2(sig_wf: np.ndarray, n_fft: int):
    """
    Zero-pad signal to the higher 2**n points for FFT
    :param sig_wf:
    :param n_fft:
    :return:
    """

    # Flatten to 2 D and memorize original shape
    n_times = sig_wf.shape[-1]

    # Legerdemain
    def _is_power_of_two(n):
        return not (n > 0 and (n & (n - 1)))

    if n_fft is None or (not _is_power_of_two(n_fft) and n_times > n_fft):
        # Compute next power of 2
        n_fft = 2 ** int(np.ceil(np.log2(n_times)))
    elif n_fft < n_times:
        raise ValueError("n_fft cannot be smaller than signal size. "
                         "Got %s < %s." % (n_fft, n_times))
    if n_times < n_fft:
        # TODO: Add verbosity
        # print('The input signal is shorter ({}) than "n_fft" ({}). '
        #       'Applying zero padding.'.format(sig_wf.shape[-1], n_fft))
        zero_pad: int = n_fft - n_times
        pad_array = np.zeros(sig_wf.shape[:-1] + (zero_pad,), sig_wf.dtype)
        sig_wf = np.concatenate((sig_wf, pad_array), axis=-1)
    else:
        zero_pad: int = 0

    return sig_wf, n_fft, zero_pad


# def tfr_stx_fft(sig_wf: np.ndarray,
#                 time_sample_interval: float,
#                 scale_order_input: float = 8.0,
#                 n_fft_in: int = None,
#                 frequency_min: float = None,
#                 frequency_max: float = None,
#                 frequency_step: float = None,
#                 factor_q: float = 0.,
#                 power_p: float = 0.,
#                 power_r: float = 1.,
#                 is_geometric: bool = False,
#                 is_inferno: bool = False,
#                 scale_base_input: float = scales.Slice.G3,
#                 scale_ref_input: float = scales.Slice.T1S,
#                 ) -> Tuple[np.ndarray, np.ndarray,  np.ndarray, np.ndarray, np.ndarray]:
#     """
#     # Stockwell transform, fft implementation.
#     # Optimized version has more free variables in sigma_scaling; testing in progress
#     :param sig_wf: input waveform. If not 2**n points, it will zero pad up
#     :param time_sample_interval: sample interval, inverse of sample rate
#     :param scale_order_input: fractional octave band order; 12 is the musical standard
#     :param n_fft_in: requested nfft, should be greater or equal to the signal length. Method zero pads up.
#     :param frequency_min: lowest stx frequency of interest
#     :param frequency_max: highest stx frequency of interest
#     :param frequency_step: stx frequency of interest if linearly sampled
#     :param factor_q: sigma_scaling adjustment, under evaluation
#     :param power_p: sigma_scaling adjustment, under evaluation
#     :param power_r: sigma_scaling adjustment. under evaluation
#     :param is_geometric: are frequencies geometrically spaced? If so, overrides frequency_step
#     :param is_inferno: are frequencies geometrically spaced and standardized as in inferno?
#     :param scale_base_input: scale base; default is base 10, base 2 octaves is scales.Slice.G2
#     :param scale_ref_input: scale reference time; default is 1 s
#     # :return: tfr_stx, psd_stx, frequency_stx_hz, frequency_stx_fft, windows_fft
#     # """
#
#     frequency_sample_rate: float = 1 / time_sample_interval
#     cycles_M: float = 12. / 5. * scale_order_input
#     lin_fft_decimate: float = 2.
#
#     # Compute the nearest higher power of two number of points for the fft
#     sig_wf_pow2, n_fft_pow2, zero_pad = sig_pad_up_to_pow2(sig_wf, n_fft_in)
#
#     # For reduction back to the near original signal after taking the ifft. Should match input sig.
#     n_fft_out = n_fft_pow2 - zero_pad
#
#     # Transformations are on zero padded signals from here onwards
#     # Take FFT and concatenate. A leaner version could let the fft do the padding.
#     sig_fft = fft(sig_wf_pow2)
#     sig_fft_cat = np.concatenate([sig_fft, sig_fft], axis=-1)
#
#     frequency_fft = fftfreq(n_fft_pow2, time_sample_interval)   # in units of 1/sample interval
#     omega_fft = 2 * np.pi * frequency_fft / frequency_sample_rate  # scaled angular frequency
#
#     window_longest_time = n_fft_pow2 / frequency_sample_rate
#     frequency_min_nth = cycles_M/window_longest_time
#
#     # Initialize stx frequencies
#     if frequency_min is None:
#         frequency_min = frequency_min_nth
#     if frequency_max is None:
#         frequency_max = frequency_sample_rate / 2.
#
#     # Computing nearest frequency later on anyway, and then using that to compute the fft.
#     start_f_idx = np.abs(frequency_fft - frequency_min).argmin()
#     stop_f_idx = np.abs(frequency_fft - frequency_max).argmin()
#     f_start = frequency_fft[start_f_idx]
#     f_stop = frequency_fft[stop_f_idx]
#
#     # Linear scale
#     if frequency_step is None:
#         # Reduce the fft resolution by a factor of lin_fft_decimate
#         frequency_step = (frequency_max - frequency_min) * lin_fft_decimate/len(frequency_fft)
#         frequency_stx = np.arange(f_start, f_stop, frequency_step)
#     else:
#         frequency_stx = np.arange(f_start, f_stop, frequency_step)
#
#     # if geometric (log) frequency
#     if is_geometric is True:
#         # if standardized to ISO3
#         if is_inferno is True:
#             order_Nth, scale_base, scale_band_number, \
#             frequency_ref, frequency_center_algebraic, frequency_center_geometric, \
#             frequency_start, frequency_end = \
#                 scales.band_frequencies_low_high(frequency_order_input=scale_order_input,
#                                                  frequency_low_input=f_start,
#                                                  frequency_high_input=f_stop,
#                                                  frequency_sample_rate_input=frequency_sample_rate,
#                                                  frequency_base_input=scale_base_input,
#                                                  frequency_ref_input=scale_ref_input)
#             frequency_stx = frequency_center_geometric
#         else:
#             num_octaves = np.log2(f_stop/f_start)
#             num_bands = int(num_octaves * scale_order_input)
#             frequency_stx = np.logspace(np.log2(f_start), np.log2(f_stop), num=num_bands, base=scale_base_input)
#
#     # Construct shifting frequency indexes
#     frequency_stx_fft = np.empty(len(frequency_stx))
#
#     # Construct time domain and fft of window
#     windows_fft = np.empty((len(frequency_stx), n_fft_pow2), dtype=np.complex128)
#     # tfr_stx_pow2 = np.empty(1, n_fft_pow2, dtype=np.complex128)
#
#     tfr_stx = np.empty((len(frequency_stx), n_fft_out), dtype=np.complex128)
#     psd_stx = np.empty((len(frequency_stx), n_fft_out))
#
#     # Minimized for loops and operations
#     for isx, fsx in enumerate(frequency_stx):
#         stx_index = np.abs(frequency_fft - fsx).argmin()
#         frequency_stx_fft[isx] = frequency_fft[stx_index]
#         # omega_sx = 2*np.pi*frequency_stx_hz/sample_rate    # non-dimensional angular stx frequency
#         omega_sx = 2 * np.pi * frequency_stx_fft[isx]/frequency_sample_rate  # eq non-dimensional angular fft frequency
#         if omega_sx == 0.:
#             windows_fft[isx] = np.ones(len(n_fft_pow2))
#         else:
#             # Sigma is the standard deviation of the Gaussian
#             # Additional flexibility is added in gamma_sigma_scaling
#             sigma_scaling = (1 + factor_q * (omega_sx ** power_p)) * (omega_sx ** (1-power_r))
#             sigma = cycles_M/omega_sx
#             sigma *= sigma_scaling
#             windows_fft[isx] = np.exp(-0.5 * (sigma ** 2.) * (omega_fft ** 2.))
#
#         # This is the main event
#         tfr_stx_pow2 = ifft(sig_fft_cat[stx_index:stx_index + n_fft_pow2] * windows_fft[isx])
#         if zero_pad > 0:
#             tfr_stx[isx, :] = tfr_stx_pow2[:-zero_pad:1]
#         else:
#             tfr_stx[isx, :] = tfr_stx_pow2
#         # Power
#         tfr_abs = np.abs(tfr_stx[isx, :])**2
#         # Add EPSILON for zero handling
#         psd_stx[isx, :] = tfr_abs + scales.EPSILON
#
#     return tfr_stx, psd_stx, frequency_stx, frequency_stx_fft, windows_fft


def stx_complex_any_scale_pow2(band_order_Nth: float,
                               sig_wf: np.ndarray,
                               frequency_sample_rate_hz: float,
                               frequency_stx_hz: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    With some assumptions and simplifications, and with some vectorization
    :param band_order_Nth: Fractional octave band - revisit
    :param sig_wf: input signal with 2^M points
    :param frequency_sample_rate_hz: sample rate in Hz
    :param frequency_stx_hz: frequency vector in increasing order
    :return: 
    """
    n_fft_pow2 = len(sig_wf)
    time_stx_s = np.arange(n_fft_pow2)/frequency_sample_rate_hz
    scale_points = len(frequency_stx_hz)
    cycles_M = scales.cycles_from_order(scale_order=band_order_Nth)
    # Take FFT and concatenate. A leaner version could let the fft do the padding.
    sig_fft = fft(sig_wf)
    sig_fft_cat = np.concatenate([sig_fft, sig_fft], axis=-1)

    frequency_fft = fftfreq(n_fft_pow2, 1/frequency_sample_rate_hz)   # in units of 1/sample interval
    omega_fft = 2 * np.pi * frequency_fft / frequency_sample_rate_hz  # scaled angular frequency
    omega_stx = 2*np.pi*frequency_stx_hz/frequency_sample_rate_hz    # non-dimensional angular stx frequency
    sigma_stx = cycles_M/omega_stx  # TODO: revisit; manage order vs frequency, use same nomenclature as cwt

    # Construct 2d matrices
    # sig_fft_cat_2d = np.tile(sig_fft_cat, (scale_points, 1))
    omega_fft_2d = np.tile(omega_fft, (scale_points, 1))
    sigma_stx_2d = np.tile(sigma_stx, (n_fft_pow2, 1)).T
    windows_fft_2d = np.exp(-0.5 * (sigma_stx_2d ** 2.) * (omega_fft_2d ** 2.))

    # Construct shifting frequency indexes
    tfr_stx = np.empty((scale_points, n_fft_pow2), dtype=np.complex128)

    # Minimal iteration
    for isx, fsx in enumerate(frequency_stx_hz):
        # This is the main event
        stx_index = np.abs(frequency_fft - fsx).argmin()
        tfr_stx[isx, :] = ifft(sig_fft_cat[stx_index:stx_index + n_fft_pow2] * windows_fft_2d[isx, :])

    return frequency_stx_hz, time_stx_s, tfr_stx


"""
This module contains unit tests for the WWZ.
"""

import unittest


class TestQuantum(unittest.TestCase):
    def test_q(self):
        """
        I don't do much yet...
        """
        self.assertTrue(True)



"""
This module calculates spectra: STFT, FFT
"""

import numpy as np
import scipy.signal as signal
from quantum_inferno import scales_dyadic as scales
from typing import Union, Tuple


# TODO: Standardize inputs
def butter_bandpass(sig_wf: np.ndarray,
                    frequency_sample_rate_hz: float,
                    frequency_cut_low_hz,
                    frequency_cut_high_hz,
                    filter_order: int = 4,
                    tukey_alpha: float = 0.5):
    """
    Butterworth bandpass filter
    :param sig_wf:
    :param frequency_sample_rate_hz:
    :param frequency_cut_low_hz:
    :param frequency_cut_high_hz:
    :param filter_order:
    :param tukey_alpha:
    :return:
    """

    nyquist = 0.5 * frequency_sample_rate_hz
    edge_low = frequency_cut_low_hz / nyquist
    edge_high = frequency_cut_high_hz / nyquist
    if edge_high >= 1:
        edge_high = 0.5  # Half of nyquist
    [b, a] = signal.butter(N=filter_order,
                           Wn=[edge_low, edge_high],
                           btype='bandpass')
    sig_taper = np.copy(sig_wf)
    sig_taper = sig_taper * signal.windows.tukey(M=len(sig_taper), alpha=tukey_alpha)
    sig_bandpass = signal.filtfilt(b, a, sig_taper)

    return sig_bandpass


def butter_highpass(sig_wf: np.ndarray,
                    frequency_sample_rate_hz: float,
                    frequency_cut_low_hz,
                    filter_order: int = 4,
                    tukey_alpha: float = 0.5):
    """
    Butterworth bandpass filter
    :param sig_wf:
    :param frequency_sample_rate_hz:
    :param frequency_cut_low_hz:
    :param filter_order:
    :param tukey_alpha:
    :return:
    """

    nyquist = 0.5 * frequency_sample_rate_hz
    edge_low = frequency_cut_low_hz / nyquist
    if edge_low >= 1:
        print('Cutoff greater than Nyquist')
        exit()
    [b, a] = signal.butter(N=filter_order,
                           Wn=[edge_low],
                           btype='highpass')
    sig_taper = np.copy(sig_wf)
    sig_taper = sig_taper * signal.windows.tukey(M=len(sig_taper), alpha=tukey_alpha)
    sig_highpass = signal.filtfilt(b, a, sig_taper)

    return sig_highpass


def butter_lowpass(sig_wf: np.ndarray,
                   frequency_sample_rate_hz: float,
                   frequency_cut_hz,
                   filter_order: int = 4,
                   tukey_alpha: float = 0.5):
    """
    Butterworth bandpass filter
    :param sig_wf:
    :param frequency_sample_rate_hz:
    :param frequency_cut_high_hz:
    :param filter_order:
    :param tukey_alpha:
    :return:
    """

    nyquist = 0.5 * frequency_sample_rate_hz
    edge_high = frequency_cut_hz / nyquist
    if edge_high >= 1:
        print('Cutoff greater than Nyquist')
        exit()
    [b, a] = signal.butter(N=filter_order,
                           Wn=[edge_high],
                           btype='lowpass')
    sig_taper = np.copy(sig_wf)
    sig_taper = sig_taper * signal.windows.tukey(M=len(sig_taper), alpha=tukey_alpha)
    sig_lowpass = signal.filtfilt(b, a, sig_taper)

    return sig_lowpass


def stft_complex_pow2(sig_wf: np.ndarray,
                      frequency_sample_rate_hz: float,
                      segment_points: int,
                      overlap_points: int = None,
                      nfft_points: int = None,
                      alpha: float = 0.25) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """ Simplest, with 50% overlap and built-in defaults """
    if nfft_points is None:
        nfft_points = int(2**np.ceil(np.log2(segment_points)))
    if overlap_points is None:
        overlap_points = int(segment_points / 2)

    frequency_stft_hz, time_stft_s, stft_complex = \
        signal.stft(x=sig_wf,
                    fs=frequency_sample_rate_hz,
                    window=('tukey', alpha),
                    nperseg=segment_points,
                    noverlap=overlap_points,
                    nfft=nfft_points,
                    detrend='constant',
                    return_onesided=True,
                    axis=-1,
                    boundary='zeros',
                    padded=True)

    return frequency_stft_hz, time_stft_s, stft_complex


def gtx_complex_pow2(sig_wf: np.ndarray,
                     frequency_sample_rate_hz: float,
                     segment_points: int,
                     gaussian_sigma: int = None,
                     overlap_points: int = None,
                     nfft_points: int = None) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

    """ Gaussian taper with 50% overlap and built-in defaults """
    if nfft_points is None:
        nfft_points = int(2**np.ceil(np.log2(segment_points)))

    # Wyatt spec is overlap_points = int(segment_points / np.sqrt(np.pi) / 8.)
    if overlap_points is None:
        overlap_points = int(segment_points / 2)

    # Gaussian window variance
    # Wyatt spec is gaussian_sigma = (segment_points + 1) / 6
    if gaussian_sigma is None:
        gaussian_sigma = int(segment_points / 4)

    frequency_stft_hz, time_stft_s, stft_complex = \
        signal.stft(x=sig_wf,
                    fs=frequency_sample_rate_hz,
                    window=('gaussian', gaussian_sigma),
                    nperseg=segment_points,
                    noverlap=overlap_points,
                    nfft=nfft_points,
                    detrend='constant',
                    return_onesided=True,
                    axis=-1,
                    boundary='zeros',
                    padded=True)

    return frequency_stft_hz, time_stft_s, stft_complex


def welch_power_pow2(sig_wf: np.ndarray,
                     frequency_sample_rate_hz: float,
                     segment_points: int,
                     nfft_points: int = None,
                     overlap_points: int = None,
                     alpha: float = 0.25) -> Tuple[np.ndarray, np.ndarray]:

    """ Simplest, with 50% overlap and built-in defaults """
    if nfft_points is None:
        nfft_points = int(2**np.ceil(np.log2(segment_points)))
    if overlap_points is None:
        overlap_points = int(segment_points / 2)
    # Compute the Welch PSD; averaged spectrum over sliding windows
    frequency_welch_hz, welch_power = \
        signal.welch(x=sig_wf,
                     fs=frequency_sample_rate_hz,
                     window=('tukey', alpha),
                     nperseg=segment_points,
                     noverlap=overlap_points,
                     nfft=nfft_points,
                     detrend='constant',
                     return_onesided=True,
                     axis=-1,
                     scaling='spectrum',
                     average='mean')
    return frequency_welch_hz, welch_power


def power_and_information_shannon_stft(stft_complex):
    """
    Computes power and information metrics
    :param stft_complex:
    :return:
    """
    power = 2 * np.abs(np.copy(stft_complex)) ** 2
    power_per_band = np.sum(power, axis=-1)
    power_per_sample = np.sum(power, axis=0)
    power_total = np.sum(power) + scales.EPSILON64
    power_scaled = power/power_total
    information_bits = -power_scaled*np.log2(power_scaled + scales.EPSILON64)
    information_bits_per_band = np.sum(information_bits, axis=-1)
    information_bits_per_sample = np.sum(information_bits, axis=0)
    information_bits_total = np.sum(information_bits) + scales.EPSILON64
    information_scaled = information_bits/information_bits_total
    return power, power_per_band, power_per_sample, power_total, power_scaled, \
           information_bits, information_bits_per_band, information_bits_per_sample, \
           information_bits_total, information_scaled


def power_and_information_shannon_welch(welch_power):
    """
    Computes power and information metrics
    :param welch_power:
    :return:
    """
    power = np.copy(welch_power)
    power_per_band = np.sum(power, axis=-1)
    power_per_sample = np.sum(power, axis=0)
    power_total = np.sum(power) + scales.EPSILON64
    power_scaled = power/power_total
    information_bits = -power_scaled*np.log2(power_scaled + scales.EPSILON64)
    information_bits_per_band = np.sum(information_bits, axis=-1)
    information_bits_per_sample = np.sum(information_bits, axis=0)
    information_bits_total = np.sum(information_bits) + scales.EPSILON64
    information_scaled = information_bits/information_bits_total
    return power, power_per_band, power_per_sample, power_total, power_scaled, \
           information_bits, information_bits_per_band, information_bits_per_sample, \
           information_bits_total, information_scaled

# def stft_from_sig(sig_wf: np.ndarray,
#                   frequency_sample_rate_hz: float,
#                   frequency_averaging_hz: float = 1.,
#                   hop_fraction: float = 0.5,
#                   tukey_alpha: float = 0.5,
#                   band_order_Nth: float = 3) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
#     """
#     Fast STFT estimate using Scipy Spectrogram. See examples s01-s04
#
#     :param sig_wf: array with input signal
#     :param frequency_sample_rate_hz: sample rate of frequency in Hz
#     :param frequency_averaging_hz: baseline lowest frequency, sets spectral resolution in FFT
#     :param hop_fraction: hop length as a fraction of the window, default of 50%
#     :param tukey_alpha: percent of the Tukey window with a cosine envelope, default of 50%
#     :param band_order_Nth: Nth order of constant Q bands, default of 3
#     :return: four numpy ndarrays with STFT, STFT_bits, time_stft_s, frequency_stft_hz
#     """
#
#     # Check
#     if hop_fraction >= 1. or hop_fraction <= 0.:
#         hop_fraction = 0.5
#     if tukey_alpha >= 1. or tukey_alpha <= 0:
#         tukey_alpha = 0.5
#
#     # NaNs, masks, etc. (if any) should be taken care of by this stage
#     sig_duration_s = len(sig_wf)/frequency_sample_rate_hz
#     _, min_frequency_hz = scales.from_duration(band_order_Nth, sig_duration_s)
#
#     if min_frequency_hz < frequency_averaging_hz:
#         # Override
#         min_frequency_hz = frequency_averaging_hz
#
#     # Compute constant Q frequency bands for the specified order for later comparison
#     order_Nth, cycles_M, quality_Q, \
#     frequency_center, frequency_start, frequency_end = \
#         scales.frequency_bands_g2f1(scale_order_input=band_order_Nth,
#                                     frequency_low_input=min_frequency_hz,
#                                     frequency_sample_rate_input=frequency_sample_rate_hz)
#
#     # Choose the spectral resolution as the key parameter
#     frequency_resolution_min_hz = np.min(frequency_end - frequency_start)
#     frequency_resolution_max_hz = np.max(frequency_end - frequency_start)
#     frequency_resolution_hz_geo = np.sqrt(frequency_resolution_min_hz*frequency_resolution_max_hz)
#     stft_time_duration_s = 1/frequency_resolution_hz_geo
#     stft_points_per_seg = int(frequency_sample_rate_hz*stft_time_duration_s)
#     # From hop fraction
#     stft_points_hop = int(hop_fraction*stft_points_per_seg)
#
#     print('STFT Duration, NFFT, HOP:', len(sig_wf), stft_points_per_seg, stft_points_hop)
#
#     # Compute the spectrogram with the spectrum/psd options
#     frequency_stft_hz, time_stft_s, sig_stft_psd_spec = \
#         signal.spectrogram(x=sig_wf,
#                            fs=frequency_sample_rate_hz,
#                            window=('tukey', tukey_alpha),
#                            nperseg=stft_points_per_seg,
#                            noverlap=stft_points_hop,
#                            nfft=stft_points_per_seg,
#                            detrend='constant',
#                            return_onesided=True,
#                            axis=-1,
#                            scaling='spectrum',
#                            mode='psd')
#
#     # Must be scaled to match scipy psd
#     stft_abs = np.sqrt(np.abs(sig_stft_psd_spec))
#     # TODO: Reconsider
#     # Unit tone amplitude -> 1/2 bit in log2 STFT amplitude (not power)
#     stft_bits = utils.log2epsilon(stft_abs)
#
#     return stft_abs, stft_bits, time_stft_s, frequency_stft_hz
#
#
# def stft_core(sig_wf: np.ndarray,
#               frequency_sample_rate_hz: float,
#               frequency_averaging_hz: float = 1.,
#               hop_fraction: float = 0.5,
#               tukey_alpha: float = 0.5,
#               band_order_Nth: float = 3,
#               ave_q_resolution = 'False') -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
#     """
#     Fast STFT estimate using Scipy stft. Baseline algorithms; potentially TMI but provides max
#     parameter control and clear documentation.
#
#     :param sig_wf: array with input signal
#     :param frequency_sample_rate_hz: sample rate of frequency in Hz
#     :param frequency_averaging_hz: baseline lowest frequency, sets spectral resolution in FFT
#     :param hop_fraction: hop length as a fraction of the window, default of 50%
#     :param tukey_alpha: percent of the Tukey window with a cosine envelope, default of 50%
#     :param band_order_Nth: Nth order of constant Q bands, default of 3
#     :return: four numpy ndarrays with STFT, STFT_bits, time_stft_s, frequency_stft_hz
#     """
#
#     # Check
#     if hop_fraction >= 1. or hop_fraction <= 0.:
#         hop_fraction = 0.5
#     if tukey_alpha >= 1. or tukey_alpha <= 0:
#         tukey_alpha = 0.5
#
#     # NaNs, masks, etc. (if any) should be taken care of by this stage
#     sig_duration_points = len(sig_wf)
#     sig_duration_s = sig_duration_points/frequency_sample_rate_hz
#     _, min_frequency_hz = scales.from_duration(band_order_Nth, sig_duration_s)
#
#     if min_frequency_hz < frequency_averaging_hz:
#         # Override
#         min_frequency_hz = frequency_averaging_hz
#
#     # Default
#     stft_time_duration_s = 1/min_frequency_hz
#
#     if ave_q_resolution:
#         # Compute constant Q frequency bands for the specified order for later comparison.
#         # Override time duration
#         order_Nth, cycles_M, quality_Q, \
#         frequency_center, frequency_start, frequency_end = \
#             scales.frequency_bands_g2f1(scale_order_input=band_order_Nth,
#                                         frequency_low_input=min_frequency_hz,
#                                         frequency_sample_rate_input=frequency_sample_rate_hz)
#
#         # Choose the spectral resolution as the key parameter
#         frequency_resolution_min_hz = np.min(frequency_end - frequency_start)
#         frequency_resolution_max_hz = np.max(frequency_end - frequency_start)
#         frequency_resolution_hz_geo = np.sqrt(frequency_resolution_min_hz*frequency_resolution_max_hz)
#         stft_time_duration_s = 1/frequency_resolution_hz_geo
#
#     stft_points_per_seg = int(frequency_sample_rate_hz*stft_time_duration_s)
#     # From hop fraction
#     stft_points_hop = int(hop_fraction*stft_points_per_seg)
#     nfft_points_per_seg = stft_points_per_seg
#     if not utils.is_power_of_two(stft_points_per_seg):
#         nfft_points_per_seg = 2**int(np.ceil(np.log2(stft_points_per_seg)))
#
#     print('STFT Duration, NPPS, HOP, NFFT:', len(sig_wf), stft_points_per_seg, stft_points_hop, nfft_points_per_seg)
#     # Compute the spectrogram with the spectrum/psd options
#     frequency_stft_hz, time_stft_s, sig_stft = \
#         signal.stft(x=sig_wf,
#                     fs=frequency_sample_rate_hz,
#                     window=('tukey', tukey_alpha),
#                     nperseg=stft_points_per_seg,
#                     noverlap=stft_points_hop,
#                     nfft=nfft_points_per_seg,
#                     detrend='constant',
#                     return_onesided=True,
#                     boundary='zeros',
#                     padded=True,
#                     axis=-1)
#
#     # Must be scaled to match scipy psd
#     stft_abs = np.abs(sig_stft)
#     # TODO: Reconsider
#     # Unit tone amplitude -> 1/2 bit in log2 STFT amplitude (not power)
#     stft_bits = utils.log2epsilon(stft_abs)
#
#     return sig_stft, stft_abs, stft_bits, time_stft_s, frequency_stft_hz
#
#
# # GENERAL FFT TOOLS
# def fft_real_dB(sig_wf: np.ndarray,
#                 sample_interval_s: float) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
#     """
#     FFT, real frequencies only, magnitude in dB. Revalidated 220620, MAG.
#
#     :param sig_wf: array with input signal
#     :param sample_interval_s: sample interval in seconds
#     :return: four numpy ndarrays with fft_frequency_pos, fft_sig_pos, fft_spectral_power_pos_dB,
#         fft_spectral_phase_radians
#     """
#     stft_fft_points = len(sig_wf)
#     fft_sig_pos = np.fft.rfft(sig_wf)
#     # returns correct RMS power level sqrt(2) -> 1
#     fft_sig_pos /= stft_fft_points
#     fft_frequency_pos = np.fft.rfftfreq(stft_fft_points, d=sample_interval_s)
#     fft_spectral_power_pos_dB = 10.*np.log10(2.*(np.abs(fft_sig_pos))**2. + EPSILON)
#     fft_spectral_phase_radians = np.angle(fft_sig_pos)
#     return fft_frequency_pos, fft_sig_pos, fft_spectral_power_pos_dB, fft_spectral_phase_radians
#
#
# def fft_real_bits(sig_wf: np.ndarray,
#                   sample_interval_s: float) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
#     """
#     FFT, real frequencies only, magnitude in bits. Revalidated 220620, MAG.
#     :param sig_wf: array with input signal
#     :param sample_interval_s: sample interval in seconds
#     :return: four numpy ndarrays with fft_frequency_pos, fft_sig_pos, fft_spectral_power_pos_bits,
#         fft_spectral_phase_radians
#     """
#     # FFT for sigetic, by the book
#     stft_fft_points = len(sig_wf)
#     fft_sig_pos = np.fft.rfft(sig_wf)
#     # returns correct RMS power level sqrt(2) -> 1
#     fft_sig_pos /= stft_fft_points
#     fft_frequency_pos = np.fft.rfftfreq(stft_fft_points, d=sample_interval_s)
#     fft_power = 2.*(np.abs(fft_sig_pos))**2.
#     # Unit tone amplitude -> 1/2 bit in log2 STFT amplitude (not power)
#     fft_spectral_power_pos_bits = utils.log2epsilon(np.sqrt(fft_power))
#     fft_spectral_phase_radians = np.angle(fft_sig_pos)
#     return fft_frequency_pos, fft_sig_pos, fft_spectral_power_pos_bits, fft_spectral_phase_radians
#
#
# def ifft_real(fft_sig_pos) -> np.ndarray:
#     """
#     Calculate the inverse of the one-dimensional discrete Fourier Transform for real input.
#
#     :param fft_sig_pos: input array
#     :return: the truncated or zero-padded input, transformed along the axis
#     """
#     ifft_sig = np.fft.irfft(fft_sig_pos).real
#     stft_fft_points = len(ifft_sig)
#     ifft_sig *= stft_fft_points
#     return ifft_sig
#
#
# def fft_complex_bits(sig_wf: np.ndarray,
#                      sample_interval_s: float) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
#     """
#     Compute the one-dimensional discrete Fourier Transform in bits
#
#     :param sig_wf: array with input signal
#     :param sample_interval_s: sample interval in seconds
#     :return: four numpy arrays with fft_frequency, fft_sig, fft_spectral_bits, fft_spectral_phase_radians
#     """
#     # FFT for sigetic, by the book
#     stft_fft_points = len(sig_wf)
#     fft_sig = np.fft.fft(sig_wf)
#     # returns correct RMS power level (not RFFT)
#     fft_sig /= stft_fft_points
#     fft_frequency = np.fft.fftfreq(stft_fft_points, d=sample_interval_s)
#     fft_spectral_bits = utils.log2epsilon(fft_sig)
#     fft_spectral_phase_radians = np.angle(fft_sig)
#     return fft_frequency, fft_sig, fft_spectral_bits, fft_spectral_phase_radians
#
#
# # Inverse Fourier Transform FFT of real function
# def ifft_complex(fft_sig_complex) -> np.ndarray:
#     """
#     Compute the one-dimensional inverse discrete Fourier Transform.
#
#     :param fft_sig_complex: input array, can be complex.
#     :return: the truncated or zero-padded input, transformed along the axis
#     """
#     ifft_sig = np.fft.ifft(fft_sig_complex)
#     stft_fft_points = len(ifft_sig)
#     ifft_sig *= stft_fft_points
#     return ifft_sig
#
#
# # Time shifted the FFT before inversion
# def fft_time_shift(fft_sig: np.ndarray,
#                    fft_frequency: np.ndarray,
#                    time_lead: Union[int, float]) -> np.ndarray:
#     """
#     Time shift an FFT. Frequency and the time shift time_lead must have consistent units and be within window
#
#     :param fft_sig: FFT signal
#     :param fft_frequency: FFT frequencies
#     :param time_lead: time shift
#     :return: numpy ndarray with time shifted FFT signal
#     """
#     # frequency and the time shift time_lead must have consistent units and be within window
#     fft_phase_time_shift = np.exp(-1j*2*np.pi*fft_frequency*time_lead)
#     fft_sig *= fft_phase_time_shift
#     return fft_sig
#
#
# # Compute Welch  from spectrogram
# def fft_welch_from_Sxx_bits(frequency_center: np.ndarray,
#                             Sxx: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
#     """
#     Compute Welch periodogram from spectrogram
#
#     :param frequency_center: array of sample frequencies.
#     :param Sxx: numpy array with spectrogram
#     :return: numpy array with frequencies, numpy array with Welch periodogram
#     """
#     # Estimate Welch periodogram by adding Sxx and dividing by the number of windows
#     # Removes zero frequency
#     Welch_Sxx = np.average(Sxx, axis=1)
#     Welch_Sxx_bits = 0.5*utils.log2epsilon(Welch_Sxx[1:])
#     f_center_nozero = frequency_center[1:]
#     return f_center_nozero, Welch_Sxx_bits
#
#
# def fft_welch_snr_power(f_center: np.ndarray,
#                         Sxx: np.ndarray,
#                         Sxx2: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
#     """
#     Calculate SNR power from specogram
#
#     :param f_center: array of sample frequencies.
#     :param Sxx: numpy array with spectogram
#     :param Sxx2: numpy array with spectogram 2. Must have the same length as Sxx
#     :return: numpy ndarray with snr frequency, numpy ndarray with snr power
#     """
#     # Estimate Welch periodogram by adding Sxx and dividing by the number of windows
#     # Removes zero frequency
#     Welch_Sxx = np.average(Sxx[1:], axis=1)
#     Welch_Sxx /= np.max(Sxx[1:])
#     Welch_Sxx2 = np.average(Sxx2[1:], axis=1)
#     Welch_Sxx2 /= np.max(Sxx2[1:])
#     snr_power = Welch_Sxx2/Welch_Sxx
#     snr_frequency = f_center[1:]
#     return snr_frequency, snr_power
#
#



"""
This module contains functions to construct quantized, standardized information packets using binary metrics.
No-chirp/sweep (index_shift=0, variable removed), simplified for the base stockwell transform.
"""

import numpy as np
import scipy.signal as signal
from quantum_inferno import scales_dyadic as scales
from typing import Tuple, Union

"""
The purpose of this code is to construct quantized, standardized information packets
using binary metrics. Based on Garces (2020). 
Cleaned up and compartmentalized for debugging
"""


def wavelet_variance_theory(amp: float, time_s: np.ndarray, scale, omega):
    """
    Theoretical variance of a Gabor wavelet
    :param amp: wavelet amplitude
    :param time_s: time support vector
    :param scale: wavelet scale
    :param omega: angular frequency
    :return: wavelet_real_var_nominal, wavelet_imag_var_nominal
    """

    wavelet_real_var_nominal = amp**2/len(time_s) * 0.5*np.sqrt(np.pi)*scale * \
                               (1 + np.exp(-(scale*omega)**2))
    wavelet_imag_var_nominal = amp**2/len(time_s) * 0.5*np.sqrt(np.pi)*scale * \
                               (1 - np.exp(-(scale*omega)**2))

    return wavelet_real_var_nominal, wavelet_imag_var_nominal


def wavelet_amplitude(scale_atom: Union[np.ndarray, float]) -> \
        Tuple[Union[np.ndarray, float], Union[np.ndarray, float]]:
    """
    Return chirp amplitude
    amp_dict_canonical = return unit integrated power and spectral energy. Good for math, ref William et al. 1991.
    amp_dict_unit_spectrum = return unit peak spectrum; for practical implementation.
    amp_dict_unity = 1. Default (no scaling), for testing and validation against real and imaginary wavelets.

    :param scale_atom: atom/logon scale
    :return: amp_canonical, amp_unit_spectrum
    """

    amp_canonical = (np.pi * scale_atom ** 2) ** (-1/4)
    amp_unit_spectrum = (4*np.pi*scale_atom**2) ** (-1/4) * amp_canonical
    return amp_canonical, amp_unit_spectrum


def amplitude_convert_norm_to_spect(scale_atom: Union[np.ndarray, float]) -> \
        Tuple[Union[np.ndarray, float], Union[np.ndarray, float]]:
    """
    Return chirp amplitude
    amp_dict_canonical = return unit integrated power and spectral energy. Good for math, ref William et al. 1991.
    amp_dict_unit_spectrum = return unit peak spectrum; for practical implementation.
    amp_dict_unity = 1. Default (no scaling), for testing and validation against real and imaginary wavelets.

    :param scale_atom: atom/logon scale
    :return: amp_canonical, amp_unit_spectrum
    """

    amp_canonical = (np.pi * scale_atom ** 2) ** (-1/4)
    amp_unit_spectrum = (4*np.pi*scale_atom**2) ** (-1/4) * amp_canonical
    amp_norm2spect = amp_unit_spectrum/amp_canonical

    return amp_norm2spect


def wavelet_time(time_s: np.ndarray,
                 offset_time_s: float,
                 frequency_sample_rate_hz: float) -> np.ndarray:
    """
    Scaled time-shifted time

    :param time_s: array with time
    :param offset_time_s: offset time in seconds
    :param frequency_sample_rate_hz: sample rate in Hz
    :return: numpy array with time-shifted time
    """
    xtime_shifted = frequency_sample_rate_hz*(time_s-offset_time_s)
    return xtime_shifted


def wavelet_complex(band_order_Nth: float,
                    time_s: np.ndarray,
                    offset_time_s: float,
                    scale_frequency_center_hz: Union[np.ndarray, float],
                    frequency_sample_rate_hz: float) -> \
        Tuple[np.ndarray, np.ndarray, Union[np.ndarray, float], Union[np.ndarray, float],
              Union[np.ndarray, float], Union[np.ndarray, float], Union[np.ndarray, float]]:

    """
    Quantized atom for specified band_order_Nth and arbitrary time duration.
    Unscaled, to be modified by the dictionary type and use case.
    Returns a frequency x time dimension wavelet vector

    :param band_order_Nth: Nth order of constant Q bands
    :param time_s: time in seconds, duration should be greater than or equal to M/fc
    :param offset_time_s: offset time in seconds, should be between min and max of time_s
    :param scale_frequency_center_hz: center frequency fc in Hz
    :param frequency_sample_rate_hz: sample rate on Hz
    :param index_shift: Redshift = -1, Blueshift = +1, None=0
    :param scale_base: G2 or G3
    :return: waveform_complex, time_shifted_s
    """

    # Center and nondimensionalize time
    xtime_shifted = wavelet_time(time_s, offset_time_s, frequency_sample_rate_hz)

    # Nondimensional chirp parameters
    scale_atom, scale_angular_frequency = \
        scales.scale_from_frequency_hz(band_order_Nth, scale_frequency_center_hz, frequency_sample_rate_hz)

    if np.isscalar(scale_atom):
        # Single frequency input
        xtime = xtime_shifted
        scale = scale_atom
        omega = scale_angular_frequency
    else:
        # Convert scale, frequency and time vectors to [frequency x time] matrices
        xtime = np.tile(xtime_shifted, (len(scale_atom), 1))
        scale = np.tile(scale_atom, (len(xtime_shifted), 1)).T
        omega = np.tile(scale_angular_frequency, (len(xtime_shifted), 1)).T

    # Base wavelet with unit absolute amplitude.
    # Note centered imaginary wavelet (sine) does not reach unity because of Gaussian envelope.
    wavelet_gabor = np.exp(-0.5*(xtime/scale)**2) * np.exp(1j*omega*xtime)
    amp_canonical, amp_unit_spectrum = wavelet_amplitude(scale)

    return wavelet_gabor, xtime_shifted, scale_angular_frequency, scale, omega, amp_canonical, amp_unit_spectrum


def wavelet_centered_4cwt(band_order_Nth: float,
                          duration_points: int,
                          scale_frequency_center_hz: Union[np.ndarray, float],
                          frequency_sample_rate_hz: float,
                          dictionary_type: str = "norm") -> \
        Tuple[np.ndarray, np.ndarray, Union[np.ndarray, float], Union[np.ndarray, float], Union[np.ndarray, float]]:

    """
    Gabor atoms for CWT computation centered on the duration of signal

    :param duration_points: number of points in the signal
    :param band_order_Nth: Nth order of constant Q bands
    :param scale_frequency_center_hz: center frequency fc in Hz
    :param frequency_sample_rate_hz: sample rate is Hz
    :param dictionary_type: Canonical unit-norm ("norm"), unit spectrum ("spect"), or unit modulus ("unit")
    :return: waveform_complex, time_shifted_s
    """

    time_s = np.arange(duration_points)/frequency_sample_rate_hz
    offset_time_s = time_s[-1]/2.

    wavelet_gabor, xtime_shifted, scale_angular_frequency, scale, omega, amp_canonical, amp_unit_spectrum = \
        wavelet_complex(band_order_Nth, time_s, offset_time_s, scale_frequency_center_hz, frequency_sample_rate_hz)

    if dictionary_type == "norm":
        amp = amp_canonical
    elif dictionary_type == "spect":
        amp = amp_unit_spectrum
    elif dictionary_type == "unit":
        if np.isscalar(scale):
            amp = 1.
        else:
            amp = np.ones(scale.shape)
    else:
        amp = amp_canonical

    wavelet_chirp = amp * wavelet_gabor
    time_centered_s = xtime_shifted/frequency_sample_rate_hz

    return wavelet_chirp, time_centered_s, scale, omega, amp


def cwt_complex_any_scale_pow2(band_order_Nth: float,
                               sig_wf: np.ndarray,
                               frequency_sample_rate_hz: float,
                               frequency_cwt_hz: np.ndarray,
                               cwt_type: str = "fft",
                               dictionary_type: str = "norm") -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Calculate CWT for chirp

    :param band_order_Nth: Nth order of constant Q bands
    :param sig_wf: array with input signal
    :param frequency_sample_rate_hz: sample rate in Hz, ordered from low to high frequency
    :param frequency_cwt_hz: center frequency vector
    :param cwt_type: one of "fft", or "morlet2". Default is "fft"
    :param index_shift: index of shift. Default is 0.0
    :param frequency_ref: reference frequency in Hz. Default is F1
    :param scale_base: G2 or G3. Default is G2
    :param dictionary_type: Canonical unit-norm ("norm") or unit spectrum ("spect"). Default is "norm"
    :return: cwt, cwt_bits, time_s, frequency_cwt_hz
    """

    wavelet_points = len(sig_wf)
    time_cwt_s = np.arange(wavelet_points)/frequency_sample_rate_hz
    scale_points = len(frequency_cwt_hz)
    cycles_M = scales.cycles_from_order(scale_order=band_order_Nth)

    cw_complex, _, _, _, amp = \
        wavelet_centered_4cwt(band_order_Nth=band_order_Nth,
                              duration_points=wavelet_points,
                              scale_frequency_center_hz=frequency_cwt_hz,
                              frequency_sample_rate_hz=frequency_sample_rate_hz,
                              dictionary_type=dictionary_type)

    if cwt_type == "morlet2":
        scale_atom, _ = \
            scales.scale_from_frequency_hz(scale_order=band_order_Nth,
                                           frequency_sample_rate_hz=frequency_sample_rate_hz,
                                           scale_frequency_center_hz=frequency_cwt_hz)
        cwt = signal.cwt(data=sig_wf, wavelet=signal.morlet2,
                         widths=scale_atom,
                         w=cycles_M,
                         dtype=np.complex128)
        if dictionary_type == 'spect':
            cwt_amp_norm2spec = amplitude_convert_norm_to_spect(scale_atom=scale_atom)
            # Convert to 2d matrix
            spec_scale = np.tile(cwt_amp_norm2spec, (wavelet_points, 1)).T
            cwt *= spec_scale

    else:
        # Convolution using the fft method
        # Convert to a 2d matrix
        sig_wf_2d = np.tile(sig_wf, (scale_points, 1))
        # Flip time
        cw_complex_fliplr = np.fliplr(cw_complex)
        cwt = signal.fftconvolve(sig_wf_2d, np.conj(cw_complex_fliplr), mode='same', axes=-1)

    return frequency_cwt_hz, time_cwt_s, cwt



import numpy as np
import scipy.signal as signal
from libquantum.scales import EPSILON
from libquantum import synthetics, utils
from typing import Optional, Tuple, Union
import matplotlib.pyplot as plt

""" Quick plotting routines """


def plot_tdr_sig(sig_wf, sig_time,
                 signal_time_base: str = 'seconds'):
    """
    Waveform
    :param sig_wf:
    :param sig_time:
    :param sig_rms_wf:
    :param sig_rms_time:
    :param signal_time_base:
    :return:
    """

    plt.figure()
    plt.plot(sig_time, sig_wf)
    plt.title('Input waveform')
    plt.xlabel("Time, " + signal_time_base)


def plot_tdr_rms(sig_wf, sig_time,
                 sig_rms_wf, sig_rms_time,
                 signal_time_base: str = 'seconds'):
    """
    Waveform
    :param sig_wf:
    :param sig_time:
    :param sig_rms_wf:
    :param sig_rms_time:
    :param signal_time_base:
    :return:
    """

    plt.figure()
    plt.plot(sig_time, sig_wf)
    plt.plot(sig_rms_time, sig_rms_wf)
    plt.title('Input waveform and RMS')
    plt.xlabel("Time, " + signal_time_base)


def plot_tfr_lin(tfr_power, tfr_frequency, tfr_time,
                 title_str: str = 'TFR, power',
                 signal_time_base: str = 'seconds'):
    """
    TFR in linear power
    :param sig_tfr:
    :param sig_tfr_frequency:
    :param sig_tfr_time:
    :param signal_time_base:
    :return:
    """

    plt.figure()
    plt.pcolormesh(tfr_time, tfr_frequency, tfr_power, cmap='RdBu_r')
    plt.title(title_str)
    plt.ylabel("Frequency, samples per " + signal_time_base)
    plt.xlabel("Time, " + signal_time_base)


def plot_tfr_bits(tfr_power, tfr_frequency, tfr_time,
                  bits_min: float = -8,
                  bits_max: float = 0,
                  title_str: str = 'TFR, top bits',
                  y_scale: str = None,
                  tfr_x_str: str = 'Time, seconds',
                  tfr_y_str: str = 'Frequency, hz',
                  tfr_y_flip: bool = False):
    """
    TFR in bits
    :param tfr_y_flip:
    :param sig_tfr:
    :param sig_tfr_frequency:
    :param sig_tfr_time:
    :param bits_max:
    :param bits_min:
    :param y_scale:
    :param title_str:
    :param tfr_x_str:
    :param tfr_y_str:
    :return: figure
    """

    tfr_bits = 0.5 * np.log2(tfr_power / np.max(tfr_power))

    fig = plt.figure()
    plt.pcolormesh(tfr_time, tfr_frequency, tfr_bits,
                   cmap='RdBu_r',
                   vmin=bits_min, vmax=bits_max, shading='nearest')
    if y_scale is None:
        plt.yscale('linear')
    else:
        plt.yscale('log')

    if tfr_y_flip:
        plt.ylim(np.max(tfr_frequency), np.min(tfr_frequency))
    plt.title(title_str)
    plt.ylabel(tfr_y_str)
    plt.xlabel(tfr_x_str)

    return fig


def plot_st_window_tdr_lin(window, freq_sx, time_fft, signal_time_base: str = 'seconds'):
    plt.figure(figsize=(8, 8))
    for j, freq in enumerate(freq_sx):
        plt.plot(time_fft, np.abs(window[j, :]), label=freq)
    plt.legend()
    plt.title('TDR window, linear')


def plot_st_window_tfr_bits(window, frequency_sx, frequency_fft, signal_time_base: str = 'seconds'):
    plt.figure(figsize=(8, 8))
    for j, freq in enumerate(frequency_sx):
        plt.plot(frequency_fft, np.log2(np.abs(window[j, :]) + EPSILON), label=freq)
    plt.legend()
    plt.title('TFR window, bits')


def plot_st_window_tfr_lin(window, frequency_sx, frequency_fft, signal_time_base: str = 'seconds'):
    plt.figure(figsize=(8, 8))
    for j, freq in enumerate(frequency_sx):
        plt.plot(frequency_fft, np.abs(window[j, :]), label=freq)
    plt.legend()
    plt.title('TFR window, lin')


def signal_gate(wf, t, tmin, tmax, fraction_cosine: float = 0):
    """
    Time gate and apply Tukey window, rectangular is the default

    :param wf: waveform
    :param t: time
    :param tmin: lower time limit
    :param tmax: upper time limit
    :param fraction_cosine: 0 = rectangular, 1 = Hann
    :return:
    """
    index_exclude = np.logical_or(t < tmin, t > tmax)
    index_include = np.logical_and(t >= tmin, t <= tmax)
    wf[index_exclude] = 0.
    wf[index_include] *= signal.windows.tukey(M=index_include.sum(), alpha=fraction_cosine)
    return wf


def oversample_time(time_duration, time_sample_interval, oversample_scale):
    """
    Return an oversampled time by a factor oversample_scale
    :param time_duration:
    :param time_sample_interval:
    :param oversample_scale:
    :return:
    """
    oversample_interval = time_sample_interval / oversample_scale
    number_points = int(time_duration / oversample_interval)
    time_all = np.arange(number_points) * oversample_interval
    return time_all


""" Reference Signatures, starting with the quantized Gabor chirp"""


def quantum_chirp(omega: float,
                  order: float = 12,
                  gamma: float = 0,
                  gauss: bool = True,
                  oversample_scale: int = 2):
    """
    Constructs a tone or a sweep with a gaussian window option and a duration of 2^n points
    :param omega: center frequency < pi. Resets to pi/4 if >= 1 (Nyquist).
    :param order: fractional octave band, sets the Q and duration. Default of 12 for chromatic scale.
    :param gamma: sweep index, could be positive or negative. Default of zero yields the atom.
    :param gauss: Apply the Gauss envelope, True as default. False constructs constant amplitude CW or sweep
    :return:
    """

    if omega >= 0.8*np.pi:
        print("Omega >= 0.8*pi (AA*Nyquist), reset to pi * 2**(-1/N")
        omega = np.pi * 2**(-1/order)

    # Gabor atom specifications
    scale_multiplier = 3 / 4 * np.pi * order

    # Atom scale
    scale = scale_multiplier / omega

    # Chirp index gamma, blueshift
    mu = np.sqrt(1 + gamma ** 2)
    chirp_scale = scale * mu

    # scale multiplier Mc
    window_support_points = 2 * np.pi * chirp_scale
    # scale up
    window_support_pow2 = 2 ** int((np.ceil(np.log2(window_support_points))))

    # Oversample by a factor of two:
    window_support_pow2_oversample = oversample_scale * window_support_pow2

    time0 = np.arange(window_support_pow2_oversample)
    time = time0 - time0[-1] / 2

    chirp_phase = omega * time + 0.5 * gamma * (time / chirp_scale) ** 2
    if gauss:
        chirp_wf_oversample = np.exp(-0.5 * (time / chirp_scale) ** 2 + 1j * chirp_phase)
    else:
        chirp_wf_oversample = np.exp(1j * chirp_phase)

    # Downsample to anti-alias
    chirp_wf = signal.decimate(x=np.real(chirp_wf_oversample), q=oversample_scale) \
               + 1j * signal.decimate(x=np.imag(chirp_wf_oversample), q=oversample_scale)

    return chirp_wf, window_support_pow2


def synth_00(frequency_0: float = 100,
             frequency_1: float = 200,
             frequency_2: float = 400,
             time_start_2: float = 0.25,
             time_stop_2: float = 0.4,
             time_sample_interval: float = 1 / 1000,
             time_duration: float = 1,
             oversample_scale: int = 2):
    """
    Generate three sine waves, oversample and decimate to AA
    Always work with  nondimensionalized units (number of points, Nyquist, etc.)

    :param frequency_0:
    :param frequency_1:
    :param frequency_2:
    :param time_start_2:
    :param time_stop_2:
    :param time_sample_interval:
    :param time_duration:
    :param oversample_scale: oversample synthetic, then decimate by scale
    :param frequency_units:
    :param time_units:
    :return:
    """

    # Oversample, then decimate
    oversample_interval = time_sample_interval / oversample_scale
    number_points = int(time_duration / oversample_interval)
    time_all = np.arange(number_points) * oversample_interval

    # Construct sine waves with unit amplitude [rms * sqrt(2)]
    sin_0 = np.sin(2 * np.pi * frequency_0 * time_all)
    signal_gate(wf=sin_0, t=time_all, tmin=0, tmax=0.5)
    sin_1 = np.sin(2 * np.pi * frequency_1 * time_all)
    signal_gate(wf=sin_1, t=time_all, tmin=0.5, tmax=1)
    sin_2 = np.sin(2 * np.pi * frequency_2 * time_all)
    signal_gate(wf=sin_2, t=time_all, tmin=time_start_2, tmax=time_stop_2)

    # Superpose gated sinusoids
    superpose = sin_0 + sin_1 + sin_2
    signal_gate(wf=superpose, t=time_all, tmin=0, tmax=1, fraction_cosine=0.05)

    # Decimate by same oversample scale, essentially an AA filter
    synth_wf = signal.decimate(x=superpose, q=oversample_scale)
    synth_time = np.arange(len(synth_wf)) * time_sample_interval

    return synth_wf, synth_time


def synth_01(a: float = 100,
             b: float = 20,
             f: float = 5,
             time_sample_interval: float = 1 / 1000,
             time_duration: float = 1,
             oversample_scale: int = 2):
    """
    Example Synthetic 1
    :param a:
    :param b:
    :param f:
    :param time_sample_interval:
    :param time_duration:
    :param oversample_scale:
    :return:
    """

    # Oversample, then decimate, essentially an AA filter
    time_all = oversample_time(time_duration, time_sample_interval, oversample_scale)
    superpose = np.cos(a * np.pi * time_all - b * np.pi * time_all * time_all) + \
                np.cos(4 * np.pi * np.sin(np.pi * f * time_all) + np.pi * 80 * time_all)
    # Taper
    signal_gate(wf=superpose, t=time_all, tmin=0, tmax=1, fraction_cosine=0.05)
    # Decimate by same oversample scale
    synth_wf = signal.decimate(x=superpose, q=oversample_scale)
    synth_time = np.arange(len(synth_wf)) * time_sample_interval

    return synth_wf, synth_time


def synth_02(t1: float = 0.3,
             t2: float = 0.7,
             t3: float = 0.5,
             f1: float = 45,
             f2: float = 75,
             f3: float = 15,
             time_sample_interval: float = 1 / 1000,
             time_duration: float = 1,
             oversample_scale: int = 2):
    """
    Example Synthetic 2
    :param t1:
    :param t2:
    :param t3:
    :param f1:
    :param f2:
    :param f3:
    :param time_sample_interval:
    :param time_duration:
    :param oversample_scale:
    :return:
    """

    t = oversample_time(time_duration, time_sample_interval, oversample_scale)

    pulse1 = np.exp(-35 * np.pi * (t - t1) ** 2) * np.cos(np.pi * f1 * t)
    pulse2 = np.exp(-35 * np.pi * (t - t2) ** 2) * np.cos(np.pi * f1 * t)
    pulse3 = np.exp(-55 * np.pi * (t - t3) ** 2) * np.cos(np.pi * f2 * t)
    pulse4 = np.exp(-45 * np.pi * (t - t3) ** 2) * np.cos(np.pi * f3 * t)

    superpose = pulse1 + pulse2 + pulse3 + pulse4

    # Decimate by same oversample scale, essentially an AA filter
    synth_wf = signal.decimate(x=superpose, q=oversample_scale)
    synth_time = np.arange(len(synth_wf)) * time_sample_interval

    return synth_wf, synth_time


def synth_03(a: float = 30,
             b: float = 40,
             c: float = 150,
             time_sample_interval: float = 1 / 1000,
             time_duration: float = 1,
             oversample_scale: int = 2):
    """

    :param a:
    :param b:
    :param c:
    :param time_sample_interval:
    :param time_duration:
    :param oversample_scale:
    :return:
    """

    # Oversample, then decimate
    time_all = oversample_time(time_duration, time_sample_interval, oversample_scale)
    superpose = np.cos(20 * np.pi * np.log(a * time_all + 1)) + \
                np.cos(b * np.pi * time_all + c * np.pi * (time_all ** 2))
    signal_gate(wf=superpose, t=time_all, tmin=0, tmax=1, fraction_cosine=0.05)

    # Decimate by same oversample scale, essentially an AA filter
    synth_wf = signal.decimate(x=superpose, q=oversample_scale)
    synth_time = np.arange(len(synth_wf)) * time_sample_interval

    return synth_wf, synth_time


""" Fancy test tone with time-domain specifications"""


def well_tempered_tone(frequency_sample_rate_hz: float = 800.,
                       frequency_center_hz: float = 60.,
                       time_duration_s: float = 16,
                       time_fft_s: float = 1.,
                       use_fft_frequency: bool = True,
                       add_noise_taper_aa: bool = False) -> Tuple[np.ndarray, np.ndarray, int, float, float, float]:
    """
    Return a tone of unit amplitude and fixed frequency
    with a constant sample rate
    :param frequency_sample_rate_hz:
    :param frequency_center_hz:
    :param time_duration_s:
    :param time_fft_s: Split the record into segments. Previous example showed 1s duration was adequate
    :param use_fft_frequency:
    :param add_noise_taper_aa:
    :return:
    """

    # The segments determine the spectral resolution
    frequency_resolution_hz = 1 / time_fft_s

    # The FFT efficiency is based on powers of 2; it is always possible to pad with zeros.
    # Set the record duration, make a power of 2. Note that int rounds down
    time_duration_nd = 2 ** (int(np.log2(time_duration_s * frequency_sample_rate_hz)))
    # Set the fft duration, make a power of 2
    time_fft_nd = 2 ** (int(np.log2(time_fft_s * frequency_sample_rate_hz)))

    # The fft frequencies are set by the duration of the fft
    # In this example we only need the positive frequencies
    frequency_fft_pos_hz = np.fft.rfftfreq(time_fft_nd, d=1 / frequency_sample_rate_hz)
    fft_index = np.argmin(np.abs(frequency_fft_pos_hz - frequency_center_hz))
    frequency_center_fft_hz = frequency_fft_pos_hz[fft_index]
    frequency_resolution_fft_hz = frequency_sample_rate_hz / time_fft_nd

    # Convert to dimensionless time and frequency, which is typically used in mathematical formulas.
    # Scale by the sample rate.
    # Dimensionless center frequency
    frequency_center = frequency_center_hz / frequency_sample_rate_hz
    frequency_center_fft = frequency_center_fft_hz / frequency_sample_rate_hz
    # Dimensionless time (samples)
    time_nd = np.arange(time_duration_nd)
    time_s = time_nd / frequency_sample_rate_hz

    if use_fft_frequency:
        # Construct synthetic tone with 2^n points and max FFT amplitude at exact fft frequency
        mic_sig = np.cos(2 * np.pi * frequency_center_fft * time_nd)
    else:
        # # Compare to synthetic tone with 2^n points and max FFT amplitude NOT at exact fft frequency
        # # It does NOT return unit amplitude (but it's close)
        mic_sig = np.cos(2 * np.pi * frequency_center * time_nd)

    if add_noise_taper_aa:
        # Add noise
        mic_sig += synthetics.white_noise_fbits(sig=mic_sig, std_bit_loss=8.)
        # Taper before AA
        mic_sig *= utils.taper_tukey(mic_sig, fraction_cosine=0.1)
        # Antialias (AA)
        synthetics.antialias_halfNyquist(mic_sig)

    print('WELL TEMPERED TONE SYNTHETIC')
    print('Nyquist frequency:', frequency_sample_rate_hz / 2)
    print('Nominal signal frequency, hz:', frequency_center_hz)
    print('FFT signal frequency, hz:', frequency_center_fft_hz)
    print('Nominal spectral resolution, hz', frequency_resolution_hz)
    print('FFT spectral resolution, hz', frequency_resolution_fft_hz)
    print('Number of signal points:', time_duration_nd)
    print('log2(points):', np.log2(time_duration_nd))
    print('Number of FFT points:', time_fft_nd)
    print('log2(FFT points):', np.log2(time_fft_nd))

    return mic_sig, time_s, time_fft_nd, frequency_sample_rate_hz, frequency_center_fft_hz, frequency_resolution_fft_hz


if __name__ == "__main__":
    sig_wf, sig_t = synth_00()
    plt.figure()
    plt.plot(sig_t, sig_wf)
    plt.title('Synth 00')

    sig_wf, sig_t = synth_01()
    plt.figure()
    plt.plot(sig_t, sig_wf)
    plt.title('Synth 01')

    sig_wf, sig_t = synth_02()
    plt.figure()
    plt.plot(sig_t, sig_wf)
    plt.title('Synth 02')

    sig_wf, sig_t = synth_03()
    plt.figure()
    plt.plot(sig_t, sig_wf)
    plt.title('Synth 03')

    plt.show()
